from package_utils.context import Context as Context_

from .config import Config
from .options import Options
from .secrets_ import Secrets

Context = Context_[Options, Config, Secrets]

context = Context(Options, Config, Secrets)



import sys
from collections.abc import Iterator

import cli

from music.models.artist import Artist
from music.runtime import runtime
from music.utils.progress import track_progress

from .artist import ArtistUpdater


def check_for_new_songs() -> None:
    description = "Checking for new songs"
    cli.console.clear_live()
    for artist in generate_artists(description):
        ArtistUpdater(artist).check_for_new_songs()


def generate_artists(description: str) -> Iterator[Artist]:  # pragma: nocover
    if runtime.is_running_in_ci:
        cli.console.print(description)
        number_of_artists = len(runtime.storage.artists)
        for i, artist in enumerate(runtime.storage.artists):
            cli.console.print(f"Checking {artist.name} ({i + 1}/{number_of_artists})")
            sys.stdout.flush()
            yield artist
    else:
        yield from track_progress(
            runtime.storage.artists,
            description=description,
            unit="artists",
        )



from music import updaters
from music.models import Path
from music.runtime import runtime


def main() -> None:
    """
    Download new songs.
    """
    if runtime.ci_context is not None:  # pragma: nocover
        with runtime.ci_context:
            _main()
    else:
        _main()


def _main() -> None:
    if runtime.context.options.clean_download_ids:
        updaters.download_ids.clean_download_ids()
    elif runtime.context.options.check_missing_downloads:
        updaters.download_ids.check_missing_downloads()
    else:
        collect_new_songs()


def collect_new_songs() -> None:
    upload_to_phone = should_upload_to_phone()
    if not upload_to_phone and not runtime.storage.tracks_ready_for_download:
        updaters.artists.check_for_new_songs()
    if runtime.storage.tracks_ready_for_download and not runtime.is_running_in_ci:
        download_new_songs()
        upload_to_phone = should_upload_to_phone()
    if upload_to_phone:
        upload_new_downloads()


def should_upload_to_phone() -> bool:
    processed_songs_present = not Path.processed_songs.is_empty()
    return runtime.context.options.upload_to_phone and processed_songs_present


def download_new_songs() -> None:
    # lazy import for performance
    from music.download import download_new_songs

    download_new_songs.download_new_songs()


def upload_new_downloads() -> None:
    # lazy import for performance
    from . import uploader

    uploader.start()



import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import Self

if TYPE_CHECKING:
    from spotdl.types.result import Result  # pragma: nocover

T = TypeVar("T", bound="Response")


@dataclass
class Response:
    name: str
    artists: tuple[str, ...]
    duration: float
    id: str

    @property
    def title(self) -> str:
        artists = ", ".join(self.artists)
        return artists + " - " + self.name

    @property
    def duration_message(self) -> str:
        duration = datetime.timedelta(seconds=self.duration)
        return str(duration).removeprefix("0:")

    @property
    def info(self) -> dict[str, str]:
        return {"title": self.title, "id": self.id, "duration": self.duration_message}

    @classmethod
    def from_result(cls, result: "Result") -> Self:
        return cls(
            artists=result.artists or (),
            name=result.name,
            duration=result.duration,
            id=result.result_id,
        )



import os
from contextlib import AbstractContextManager
from functools import cached_property
from typing import TYPE_CHECKING

from music.clients import spotify
from music.context import Context, context
from music.storage.storage import Storage

if TYPE_CHECKING:
    from spotdl.providers.audio import YouTubeMusic  # pragma: nocover


class Runtime:
    def __init__(self) -> None:
        self.context: Context = context

    @cached_property
    def spotify_client(self) -> spotify.Client:
        return spotify.Client(self.context.secrets)

    @cached_property
    def youtube_music_client(self) -> "YouTubeMusic":
        # slow import
        from spotdl.providers.audio.ytmusic import YouTubeMusic

        return YouTubeMusic()

    @cached_property
    def storage(self) -> Storage:
        return Storage()

    @cached_property
    def is_running_in_ci(self) -> bool:
        return (
            "GITHUB_ACTIONS" in os.environ and "PYTEST_CURRENT_TEST" not in os.environ
        )

    @cached_property
    def ci_context(self) -> AbstractContextManager[None] | None:
        ci_context = None
        if context.is_running_in_ci:  # pragma: nocover
            # imports optional dependencies
            try:
                from music.utils.ci_context import CIContext
            except ImportError:
                pass
            else:
                ci_context = CIContext()
        return ci_context


runtime = Runtime()



from collections.abc import Iterator

from music.models.path import Path
from music.runtime import runtime


def clean_download_ids() -> None:
    track_info_iterator = generate_tracks_with_unique_name()
    ids, names = zip(*track_info_iterator, strict=False)
    tracks = runtime.spotify_client.songs(ids)
    downloaded_tracks = {
        track.id: name for track, name in zip(tracks, names, strict=False)
    }
    runtime.storage.downloaded_tracks = downloaded_tracks


def generate_tracks_with_unique_name() -> Iterator[tuple[str, str]]:
    encountered_names = set()
    for id_, name in runtime.storage.downloaded_tracks.items():
        if name not in encountered_names:
            encountered_names.add(name)
            yield id_, name


def check_missing_downloads() -> None:
    directories = Path.all_songs, Path.deleted
    downloaded_names = {
        path.stem for directory in directories for path in directory.iterdir()
    }
    runtime.storage.tracks_to_download = {
        id_: name
        for id_, name in runtime.storage.downloaded_tracks.items()
        if name.replace("?", "") not in downloaded_names
    }



from dataclasses import dataclass
from typing import Annotated

import typer

from music.models.path import Path


class Helps:
    clean_download_ids = (
        "Spotify changes the ids of its tracks from time to time.\n"
        "This command saves the most recent id of all downloaded tracks"
    )
    check_missing_downloads = (
        "Check downloaded ids to be actually present as downloaded "
        "songs and add missing ids to download queue"
    )


class TyperOptions:
    clean_download_ids = typer.Option(help=Helps.clean_download_ids)
    check_missing_downloads = typer.Option(help=Helps.check_missing_downloads)


@dataclass
class Options:
    config_path: Path = Path.config
    upload_to_phone: bool = True
    clean_download_ids: Annotated[bool, TyperOptions.clean_download_ids] = False
    check_missing_downloads: Annotated[bool, TyperOptions.check_missing_downloads] = (
        False
    )
    fix_mtimes_on_phone: bool = False
